
/* pca200e.c: A FORE Systems PCA-200E driver for ATM on Linux. */
/*  $Id: pca200e.c,v 1.9 1998/01/05 18:42:17 ud3 Exp $ */

/*
    Written 1996-97 by Uwe Dannowski.

    This software may be used and distributed according to the terms
    of the GNU Public License, incorporated herein by reference.

    The author may be reached as ud3@inf.tu-dresden.de

    This driver should work with the FORE Systems PCA-200E.
    It won't work with the PCA-200 or with the ForeRunner LE.

*/


#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/bios32.h>
#include <linux/pci.h>
#include <linux/malloc.h>
#include <linux/atmdev.h>
#include <linux/sonet.h>
#include <asm/io.h>
#include <asm/string.h>
#include <asm/segment.h>

#include <linux/delay.h>

#ifdef MODULE
#include <linux/module.h>
#endif

#include "pca200e.h"

/*
  CONFIG_ATM_PCA200E_DEBUG level intention:	0 no debug info at all
					       10 full debug info
 */

// #define CONFIG_ATM_PCA200E_DEBUG 4

/* if CONFIG_ATM_PCA200E_SYNC_SEND is defined pca200e_send() polls
   until the PDU has been sent or error */
// #define CONFIG_ATM_PCA200E_SYNC_SEND



#ifdef CONFIG_ATM_PCA200E_DEBUG
#define DPRINTK(format,args...) printk(DEV_LABEL format,##args)
#else
#define DPRINTK(format,args...)
#endif

#if CONFIG_ATM_PCA200E_DEBUG >= 8
#define TRACE printk(DEV_LABEL __FILE__": %d: "__PRETTY_FUNCTION__"\n", __LINE__);
#else
#define TRACE
#endif

/* support old naming sillyness */
#if LINUX_VERSION_CODE < 0x020100                                     
#define ioremap vremap
#define iounmap vfree
#endif


/* references to the firmware image */
extern const firmheader_t	_binary_pca200e_bin_start[];
extern const unsigned      	_binary_pca200e_bin_size;

/* pointer to the list of physical devices */
struct pca200e_dev *	pca200e_boards		= NULL;

/* this holds the available cell rate for non-UBR connections,
   regardless of what UBR connections use up */
//static int available_cell_rate = ATM_OC3_PCR;


/* resolve discrepancy between ATM_AALx values and AALx numbers */
inline unsigned char
aal2def(unsigned char aal)
{
  switch (aal)
    {
    case 0: return ATM_AAL0;
    case 4: return ATM_AAL34;
    case 5: return ATM_AAL5;
    };
  return ATM_NO_AAL;
};

inline unsigned char
def2aal(unsigned char def)
{
  switch (def)
    {
    case ATM_AAL0: return 0;
    case ATM_AAL1: return 1;
    case ATM_AAL2: return 2;
    case ATM_AAL34: return 4;
    case ATM_AAL5: return 5;
    };
  return -1;
};

int
pca200e_card_detect(struct pca200e_dev * pca200e, unsigned index)
{
  unsigned short pci_command;

  if (!pcibios_present())
    {
      printk("PCA-200E driver but no PCI BIOS ???\n");
      return -ENODEV;
    };
  
  if(pcibios_find_device(PCI_VENDOR_ID_FORE,
			 PCI_DEVICE_ID_FORE_PCA200E,
                         index,
			 &pca200e->bus,
			 &pca200e->dev_fn))
    return -ENODEV;
  
  if (pcibios_read_config_dword(pca200e->bus,
				pca200e->dev_fn,
				PCI_BASE_ADDRESS_0,
                                (unsigned*) &pca200e->phys_rambase) != PCIBIOS_SUCCESSFUL)
    {
      printk("Can't read membase\n");
      return -EIO;
    };
  
  if (pcibios_read_config_byte(pca200e->bus,
			       pca200e->dev_fn,
			       PCI_INTERRUPT_LINE,
                               (unsigned char*) &pca200e->irq) != PCIBIOS_SUCCESSFUL)
    {
      printk("Can't read IRQ\n");
      return -EIO;
    };

   if (pcibios_write_config_byte(pca200e->bus,
				pca200e->dev_fn,
				PCI_LATENCY_TIMER, 
				255) != PCIBIOS_SUCCESSFUL)
    {
      printk("Can't write PCI_LATENCY_TIMER\n");
      return -EIO;
    };
  
  
  if (pcibios_read_config_word(pca200e->bus,
			       pca200e->dev_fn,
			       PCI_COMMAND,
                               &pci_command) != PCIBIOS_SUCCESSFUL)
    {
      printk("Can't read PCI_COMMAND\n");
      return -EIO;
    };
  
  pci_command |= (  PCI_COMMAND_MEMORY
		  | PCI_COMMAND_SPECIAL
		  | PCI_COMMAND_MASTER
		  | PCI_COMMAND_INVALIDATE
		  );

  pci_command &= ~(PCI_COMMAND_IO);
  pci_command = -1;

  if (pcibios_write_config_word(pca200e->bus,
				pca200e->dev_fn,
				PCI_COMMAND, 
				pci_command) != PCIBIOS_SUCCESSFUL)
    {
      printk("Can't write PCI_COMMAND\n");
      return -EIO;
    };

  return 0;
};



int
pca200e_map(struct pca200e_dev* pca200e)
{
#ifdef __i386__
  pca200e->virt_rambase = (unsigned long) ioremap(pca200e->phys_rambase,
						  PCA200E_HARDWARE_BASE_LENGTH);
#endif /* __i386__ */
#ifdef __alpha__
  pca200e->virt_rambase = pca200e->phys_rambase;
#endif /* __alpha__ */

  if (!pca200e->virt_rambase)
    {
      printk(DEV_LABEL "ioremap: can't set up mapping\n");
      return -EFAULT;
    }

  pca200e->cardstate = cs_map;
  return 0;
};

int
pca200e_reset(struct pca200e_dev* pca200e)
{
  int i;
  monitor_t* m   = (monitor_t*) (pca200e->virt_rambase + PCA200E_MON960_OFFSET);

  /* write COLD_START in boot status */
  writel(cold_start, &m->bstat);

  /* flash the "Board Reset Bit" in HCR */
  writeb(0x01, pca200e->virt_rambase + PCA200E_HCR_OFFSET);
  i=jiffies; while (jiffies <= i+MSECS(10)); /* 10..20 ms	*/
  writeb(0x00, pca200e->virt_rambase + PCA200E_HCR_OFFSET);

  /* wait for selftest to complete */
  i = POLL_STATUS(readl(&m->bstat), self_test_ok, PCA200E_INIT_TIMEOUT*10);
  if (!i)
    {
      printk(DEV_LABEL "self test FAILED.\n");
      return -EIO;
    };
  
  pca200e->cardstate = cs_reset;
  return 0;
};

void
pca200e_loadfirm(struct pca200e_dev* pca200e)
{
  /* copy the firmware image into shared memory -
     the load address is given in the firmware header */
  
#ifdef __i386__
  memcpy_toio(pca200e->virt_rambase + _binary_pca200e_bin_start->loadoffset,
	      _binary_pca200e_bin_start,
	      (unsigned) &_binary_pca200e_bin_size);
#endif /* __i386__ */
#ifdef __alpha__
  /* broken memcpy_toio - PCA-200E allows 32 bit accesses only */
  unsigned* d = (unsigned*) (pca200e->virt_rambase + _binary_pca200e_bin_start->loadoffset);
  unsigned* s = (unsigned*) _binary_pca200e_bin_start;
  unsigned  i = (unsigned)  &_binary_pca200e_bin_size / sizeof(unsigned) + 1;
  
  for (;i--;s++,d++)
    writel(*s, d);
#endif /* __alpha__ */

  pca200e->cardstate = cs_loadfirm;
};

void
pca200e_mon960put_c(struct pca200e_dev* pca200e, unsigned char c)
{
  monitor_t * m = (monitor_t *)(pca200e->virt_rambase+PCA200E_MON960_OFFSET);
  unsigned long i;
  
  /* read and forget all chars from mon960. we have to do this because
     mon960 doesnt accept any new chars if it has someting to "say" */
  i = jiffies;
  while(jiffies < i+MSECS(20))	/* 20 ms */
    if (readl(&m->uart.recv) & PCA200E_MON960_UART_AVAIL)
      {
//	printk("%c", readl(&m->uart.recv));
	writel(PCA200E_MON960_UART_FREE, &m->uart.recv);
	i=jiffies;
      };
  
  /* send the character to mon960 */
  writel(c | PCA200E_MON960_UART_AVAIL, &m->uart.send);
  
  /* read and forget all chars being sent from mon960 as a reply */
  i = jiffies;
  while(jiffies < i+MSECS(20))	/* 20 ms */
    if (readl(&m->uart.recv) & PCA200E_MON960_UART_AVAIL)
      {
//	printk("%c", readl(&m->uart.recv));
	writel(PCA200E_MON960_UART_FREE, &m->uart.recv);
	i=jiffies;
      };
};


unsigned int
pca200e_startfirm(struct pca200e_dev* pca200e)
{
  char s[16+32];
  int i;
  monitor_t* m   = (monitor_t*)(pca200e->virt_rambase + PCA200E_MON960_OFFSET);

  /* issue the "go <address>" command to the mon960 
     the start address is given in the firmware header */
  sprintf(s,"\rgo %x\r", _binary_pca200e_bin_start->startoffset);

  for (i=0; s[i] != 0x00; i++)
    pca200e_mon960put_c(pca200e, s[i]);

  /* wait for firmware getting started */
  i = POLL_STATUS(readl(&m->bstat), cp_running, PCA200E_INIT_TIMEOUT);
  if (!i)
    {
      printk(DEV_LABEL "firmware DIDN'T START.\n");
      return -EIO;
    };
  writel(1, &pca200e->cpqueue->imask);
  pca200e->cardstate = cs_startfirm;
  return 0;
};

#define SET_QUEUE_PARAM(q,qs,bs,ps,sb) \
writel(qs, &q.queue_length); \
writel(bs, &q.buffer_size); \
writel(ps, &q.pool_size); \
writel(sb, &q.supply_blksize);

unsigned int
pca200e_initialize(struct pca200e_dev* pca200e)
{
  int i,j;

  pca200e->cpqueue = (cp_queues_t*) (pca200e->virt_rambase + PCA200E_COMMON_ORIGIN);

  /* fill the init command parameters */
  writel(290,			&pca200e->cpqueue->init.receive_threshold);
  writel(NUM_CONNECTIONS,	&pca200e->cpqueue->init.num_connect);
  writel(QSIZE_CMD,		&pca200e->cpqueue->init.cqueue_len);
  writel(QSIZE_RX,		&pca200e->cpqueue->init.rqueue_len);
  writel(QSIZE_TX,		&pca200e->cpqueue->init.tqueue_len);
  writel(0x0E,			&pca200e->cpqueue->init.conless_vpvc);
  writel(RPD_EXTENSION,		&pca200e->cpqueue->init.rpd_extension);
  writel(TPD_EXTENSION,		&pca200e->cpqueue->init.tpd_extension);

  SET_QUEUE_PARAM(pca200e->cpqueue->init.small_b1,
		  QSIZE_RESUPPLY, SMALL_B1_BUFF_SIZE,
		  SMALL_B1_POOL_SIZE, RBD_BLK_SIZE);
  SET_QUEUE_PARAM(pca200e->cpqueue->init.large_b1,
		  QSIZE_RESUPPLY, LARGE_B1_BUFF_SIZE,
		  LARGE_B1_POOL_SIZE, RBD_BLK_SIZE);
  SET_QUEUE_PARAM(pca200e->cpqueue->init.small_b2,
		  0, 0, 0, 0);
  SET_QUEUE_PARAM(pca200e->cpqueue->init.large_b2,
		  0, 0, 0, 0);

  /* issue the init command */
  writel(stat_pending,	&pca200e->cpqueue->init.status);
  writel(op_initialize,	&pca200e->cpqueue->init.op);

  /* wait for init command to complete */
  i = POLL_STATUS(readl(&pca200e->cpqueue->init.status), stat_complete,
		  PCA200E_INIT_TIMEOUT);
  if (!i)
    {
      printk(DEV_LABEL "init command failed!!!\n");
      return -EIO;
    };
  
  /* check if heartbeat is counting upwards - this is pure luxury */
  i = jiffies; j = readl(&pca200e->cpqueue->heartbeat);
  while (jiffies < i+ 10*PCA200E_INIT_TIMEOUT)
    if (j < readl(&pca200e->cpqueue->heartbeat))
      break;
    else
      j = readl(&pca200e->cpqueue->heartbeat);

  if (j >= readl(&pca200e->cpqueue->heartbeat))
    {
      printk(DEV_LABEL "heartbeat not counting upwards!!!\n");
      return -EIO;
    };

  pca200e->cardstate = cs_init;
  return 0;
};

int
pca200e_initcmdqueue(struct pca200e_dev* pca200e)
{
  int i;
  qhost_cmd_t *tmp1, *tmp2 = NULL;
  
  pca200e->cmd_status_page = __get_dma_pages(GFP_KERNEL,
				      (QSIZE_CMD*32+(PAGE_SIZE-1))/PAGE_SIZE);
  if (!(pca200e->cmd_status_page))
    {
      printk(DEV_LABEL "initcmdqueue: cmd_status_page: get_dma_pages failed\n");
      return -ENOMEM;
    };
  for (i=0; i<QSIZE_CMD; i++)
    {
      tmp1 = kmalloc(sizeof(qhost_cmd_t), GFP_KERNEL);
      if (!tmp1)
	{
	  printk(DEV_LABEL "initcmdqueue: kmalloc failed - index=%d\n",i);
	  return -ENOMEM;
	};
      memset((void*) tmp1, 0, sizeof(qhost_cmd_t));
      if(i)
	tmp2->next = tmp1;
      else
	pca200e->cmd_queue = tmp1;

      tmp1->qentry = (qcard_cmd_t*) (pca200e->virt_rambase
				     +readl(&pca200e->cpqueue->command_queue)
				     +i*sizeof(qcard_cmd_t));
      tmp1->status = (fw_status_t*) (pca200e->cmd_status_page+i*32);
// printk("\ntmp1->status: %p, bus: %lx, phys: %lx", tmp1->status, virt_to_bus(tmp1->status), virt_to_bus(tmp1->status));
      writel(virt_to_bus(tmp1->status), &tmp1->qentry->stataddr);
      SET_STATUS(tmp1, stat_free);
      tmp2 = tmp1;
    };
  tmp1->next=pca200e->cmd_queue;

  pca200e->cardstate = cs_init_cmd;
  return 0;
};

int
pca200e_inittxqueue(struct pca200e_dev* pca200e)
{
  int i;
  qhost_tx_t *tmp1, *tmp2=0;
  
  pca200e->tpd_page = __get_dma_pages(GFP_KERNEL,
				      (QSIZE_TX*TPD_SIZE+(PAGE_SIZE-1))/(PAGE_SIZE));
  if (!pca200e->tpd_page)
    {
      printk(DEV_LABEL "inittxqueue: tpd_page: get_dma_pages failed\n");
      return -ENOMEM;
    };
  pca200e->tx_status_page = __get_dma_pages(GFP_KERNEL,
					    (QSIZE_TX*32+(PAGE_SIZE-1))/PAGE_SIZE);
  if (!pca200e->tx_status_page)
    {
      printk(DEV_LABEL "inittxqueue: tx_status_page: get_dma_pages failed\n");
      return -ENOMEM;
    };
  for (i=0; i<QSIZE_TX; i++)
    {
      tmp1 = kmalloc(sizeof(qhost_tx_t), GFP_KERNEL);
      if (!tmp1)
	{
	  printk(DEV_LABEL "inittxqueue: kmalloc failed - index=%d\n",i);
	  return -ENOMEM;
	};
      memset((void*) tmp1, 0, sizeof(qhost_tx_t));
      if (i)
	tmp2->next = tmp1;
      else
	pca200e->tx_queue = tmp1;

      tmp1->qentry = (qcard_t*) (pca200e->virt_rambase
				 +readl(&pca200e->cpqueue->transmit_queue)
				 +i*sizeof(qcard_t));
      tmp1->status = (fw_status_t*)(pca200e->tx_status_page+i*32);
      writel(virt_to_bus(tmp1->status), &tmp1->qentry->stataddr);
      SET_STATUS(tmp1, stat_free);
      tmp1->tpd	= (tpd_t*)(pca200e->tpd_page+i*TPD_SIZE);
      tmp2 = tmp1;
    };
  tmp1->next = pca200e->tx_queue;
  pca200e->tx_queue_tail = pca200e->tx_queue;

  pca200e->cardstate = cs_init_tx;
  return 0;
};

int
pca200e_initrxqueue(struct pca200e_dev* pca200e)
{
  int i;
  qhost_rx_t *tmp1, *tmp2 = NULL;
  
  pca200e->rpd_page = __get_dma_pages(GFP_KERNEL,
				      (QSIZE_RX*RPD_SIZE+(PAGE_SIZE-1))/PAGE_SIZE);
  if (!pca200e->rpd_page)
    {
      printk(DEV_LABEL "initrxqueue: rpd_page: get_dma_pages failed\n");
      return -ENOMEM;
    };
  pca200e->rx_status_page = __get_dma_pages(GFP_KERNEL,
					    (32*QSIZE_RX+(PAGE_SIZE-1))/(PAGE_SIZE));
  if (!pca200e->rx_status_page)
    {
      printk(DEV_LABEL "initrxqueue: rx_status: get_dma_pages failed\n");
      return -ENOMEM;
    };
  for (i=0; i<QSIZE_RX; i++)
    {
      tmp1 = kmalloc(sizeof(qhost_rx_t), GFP_KERNEL);
      if (!tmp1)
	{
	  printk(DEV_LABEL "initrxqueue: kmalloc failed - index=%d\n",i);
	  return -ENOMEM;
	};
      memset((void*) tmp1, 0, sizeof(qhost_rx_t));
      if(i)
	tmp2->next = tmp1;
      else
	pca200e->rx_queue = tmp1;
      
      tmp1->qentry = (qcard_t*) (pca200e->virt_rambase
				 +readl(&pca200e->cpqueue->receive_queue)
				 +i*sizeof(qcard_t));
      tmp1->status = (fw_status_t*) (pca200e->rx_status_page+i*32);
      writel(virt_to_bus(tmp1->status), &tmp1->qentry->stataddr);
      SET_STATUS(tmp1, stat_free);
      tmp1->rpd	= (rpd_t*)(pca200e->rpd_page+i*RPD_SIZE);
      writel(virt_to_bus((void*) tmp1->rpd), &tmp1->qentry->ioblock);
      tmp2 = tmp1;
    };
  tmp1->next = pca200e->rx_queue;
  
  pca200e->cardstate = cs_init_rx;
  return 0;
};

int
pca200e_buildrbuffqueue(struct pca200e_dev* pca200e, int sml)
{
  unsigned i;
  qhost_rbuff_t* tmp1 = NULL;
  qhost_rbuff_t* tmp2 = NULL;
  
  qhost_rbuff_t** ptr;		/* pointer to queue pointer	*/
  unsigned int count;		/* number of buffers		*/
  unsigned int size;		/* buffer size			*/
  buffer_magn_t mag;		/* buffer magnitude		*/
  
  ptr = (sml ? &pca200e->smlb1_rbuff_head : &pca200e->lrgb1_rbuff_head);
  count = (sml ? SMALL_B1_NUM_RBUFFS : LARGE_B1_NUM_RBUFFS);
  size = (sml ? SMALL_B1_BUFF_SIZE : LARGE_B1_BUFF_SIZE);
  mag = (sml ? bmag_small : bmag_large);
  
  for (i = 0; i < count; i++)
    {
      tmp1 = kmalloc(sizeof(qhost_rbuff_t), GFP_KERNEL);
      if (!tmp1)
	{
	  printk(DEV_LABEL "buildrbuffqueue: kmalloc failed - index=%d\n",i);
	  *ptr = tmp2;
	  return -ENOMEM;
	};
      pca200e->bufferlist[((unsigned long) (pca200e->bufferlist[0]))++] = tmp1;
      memset((void*) tmp1, 0, sizeof(qhost_rbuff_t));
      tmp1->type = mag;    
      tmp1->next = tmp2;
      tmp1->buff = kmalloc(size, GFP_KERNEL);
      if (!tmp1->buff)
	{
	  printk(DEV_LABEL "buildrbuffqueue: buffer kmalloc failed - index=%d\n",i);
	  *ptr = tmp1;
	  return -ENOMEM;
	};
      tmp2 = tmp1;
    };
  *ptr = tmp1;
  return 0;
};

void
pca200e_supply(struct pca200e_dev* pca200e, int sml)
{
  int i;
  
  qhost_supp_t**  sptr;
  qhost_rbuff_t** rptr;
  int* cptr;

  sptr = (sml ? &pca200e->smlb1_queue : &pca200e->lrgb1_queue);
  rptr = (sml ? &pca200e->smlb1_rbuff_head : &pca200e->lrgb1_rbuff_head);
  cptr = (sml ? &pca200e->smlb1_count : &pca200e->lrgb1_count);

  for (i=0; i<RBD_BLK_SIZE; i++)
    {
      (*sptr)->rbd[i].buffer = virt_to_bus((*rptr)->buff);
      (*sptr)->rbd[i].handle = virt_to_bus((void*) *rptr);
      *rptr = (*rptr)->next;
    };

  SET_STATUS((*sptr), stat_pending);
  writel((unsigned long) (*sptr)->rbd, &((*sptr)->qentry->ioblock));
  *sptr = (*sptr)->next;
  *cptr += RBD_BLK_SIZE;
};

int
pca200e_initbuffqueue(struct pca200e_dev* pca200e, int sml)
{
  int i;
  qhost_supp_t*		tmp1 = NULL;
  qhost_supp_t*		tmp2 = NULL;
  unsigned long*	stpptr;
  unsigned long*	rbpptr;
  qhost_supp_t**	qptr;
  unsigned long		cpq;
  unsigned*		cptr;
  enum cardstate_t	state;

  stpptr = (sml ? &pca200e->smlb1_status_page : &pca200e->lrgb1_status_page);
  rbpptr = (sml ? &pca200e->smlb1_rbd_page : &pca200e->lrgb1_rbd_page);
  qptr = (sml ? &pca200e->smlb1_queue : &pca200e->lrgb1_queue);
  cpq = readl((sml ? &pca200e->cpqueue->small_b1_queue : &pca200e->cpqueue->large_b1_queue));
  cptr = (sml ? &pca200e->smlb1_count : &pca200e->lrgb1_count);
  state = (sml ? cs_init_sb1 : cs_init_lb1);

  *stpptr = __get_dma_pages(GFP_KERNEL, (32*QSIZE_RESUPPLY+(PAGE_SIZE-1))/PAGE_SIZE);
  if (!*stpptr)
    {
      printk(DEV_LABEL "initbuffqueue: status_page: get_dma_pages failed\n");
      return -ENOMEM;
    };
  *rbpptr = __get_dma_pages(GFP_KERNEL, (QSIZE_RESUPPLY*RBD_BLK_SIZE*sizeof(rbd_t)+4095)/4096);
  if (!*rbpptr)
    {
      printk(DEV_LABEL "initbuffqueue: rbd_page: get_dma_pages failed\n");
      return -ENOMEM;
    };
  for (i=0; i<QSIZE_RESUPPLY; i++)
    {
      tmp1 = kmalloc(sizeof(qhost_supp_t), GFP_KERNEL);
      if (!tmp1)
	{
	  printk(DEV_LABEL "initbuffqueue: kmalloc failed - index=%d\n",i);
	  return -ENOMEM;
	};
      memset((void*) tmp1, 0, sizeof(qhost_supp_t));
      if (i)
	tmp2->next = tmp1;
      else
	*qptr = tmp1;
      
      tmp1->qentry = (qcard_t*) (pca200e->virt_rambase+cpq+i*sizeof(qcard_t));
      tmp1->status = (fw_status_t*) (*stpptr + i*32);
      writel(virt_to_bus(tmp1->status), &tmp1->qentry->stataddr);
      SET_STATUS(tmp1, stat_free);
      tmp1->rbd = (rbd_t*)(*rbpptr+i*RBD_BLK_SIZE*sizeof(rbd_t));
      tmp2 = tmp1;
    };
  tmp1->next = *qptr;
  *cptr = 0;
  
  if (pca200e_buildrbuffqueue(pca200e, sml))
    return -ENOMEM;
  pca200e_supply(pca200e, sml);
  
  pca200e->cardstate = state;
  return 0;
};

/* read the card PROM - mainly to get ESI */
int
pca200e_readprom(struct pca200e_dev* pca200e)
{
  unsigned i;
  expansion_rom_data_t* exp;

  exp = (expansion_rom_data_t*) __get_dma_pages(GFP_KERNEL,1);
  if (!exp)
    return -ENOMEM;

  memset((void*) exp, 0, sizeof(expansion_rom_data_t));

  writel(virt_to_bus((void*)exp),
	 &pca200e->cmd_queue->qentry->cmd.prom_data_param.prom_data_buff);
  SET_STATUS(pca200e->cmd_queue, stat_pending);
  writel(op_get_prom_data,
	 &pca200e->cmd_queue->qentry->cmd.prom_data_param.op);

  i = POLL_STAT_COMPLETE(pca200e->cmd_queue, PCA200E_COMMAND_TIMEOUT*1000);

  if (i) {
    printk(DEV_LABEL "readprom command failed %x!!!\n", i);
    return -EBUSY;
  };

  printk(DEV_LABEL "Rev %c, S/N: %d, ESI: %02x:%02x:%02x:%02x:%02x:%02x\n",
	 (char)((exp->hw_version & 0xff)+'@'),
	 exp->serial_number,
	 exp->mac_addr[2], exp->mac_addr[3], exp->mac_addr[4],
	 exp->mac_addr[5], exp->mac_addr[6], exp->mac_addr[7]);
  
  for (i=2;i<8;i++)
    pca200e->mac[i-2] = exp->mac_addr[i];
  
  SET_STATUS(pca200e->cmd_queue, stat_free);
  pca200e->cmd_queue = pca200e->cmd_queue->next;
  
  free_pages((long)exp,1);
  return 0;
};

void
pca200e_getstats(struct pca200e_dev* pca200e, stats200_t* stats)
{
  int i;
  qhost_cmd_t*	cq;

  cq = pca200e->cmd_queue;
  writel(virt_to_bus((void*) stats), &cq->qentry->cmd.stats_param.stats_buff);
  SET_STATUS(cq, stat_pending);
  writel(op_request_stats, &cq->qentry->cmd.stats_param.op);
  
  i = POLL_STAT_COMPLETE(cq, PCA200E_COMMAND_TIMEOUT);
  if (i)
    printk(DEV_LABEL "getstats command failed!!!\n");
  
  SET_STATUS(cq, stat_free);
  pca200e->cmd_queue = cq->next;
};

void
pca200e_inttx(struct atm_dev* dev)
{
  qhost_tx_t*	txq;
  int	i;
  
  txq = PCA200E_DEV(dev)->tx_queue_tail;
  for (i=0; i<QSIZE_TX; i++)
    {
      if (STAT_COMPLETE(txq))
	{
	  if (((unsigned long) txq->skb->data) & 0x00000003)
	    kfree(txq->txbuff);
	  if (txq->vcc->pop)
	    txq->vcc->pop(txq->vcc, txq->skb);
	  else
	    dev_kfree_skb(txq->skb, FREE_WRITE);
	  if (STAT_ERROR(txq))
	    txq->vcc->stats->tx_err++;
	  else
	    txq->vcc->stats->tx++;
	  SET_STATUS(txq, stat_free);
	}
      txq = txq->next;
#if 0
      else
	break;
      PCA200E_DEV(dev)->tx_queue_tail = PCA200E_DEV(dev)->tx_queue_tail->next;
#endif
    };
};

void
pca200e_intrx(struct atm_dev* dev)
{
  qhost_rx_t*	rxq;
  struct pca200e_dev* pca200e;
  int j;
  rpd_t* rpd;
  struct atm_vcc* vcc;
 
  pca200e = PCA200E_DEV(dev);
  rxq = pca200e->rx_queue;

  while(1) // was:  for (i=0; i<QSIZE_RX*2; i++)
    {
      /* check if entry has been completed */
      if (STAT_COMPLETE(rxq))
	{
	  rpd = rxq->rpd;
	  
	  /* check if entry is ok */
	  if (!STAT_ERROR(rxq))
	    {
	      /* walk through the list of open vccs to find the right one */
	      for (vcc = dev->vccs; vcc; vcc = vcc->next)
		if (((vcc->vpi) == (rpd->atm_header.vpi)) &&
		    ((vcc->vci) == (rpd->atm_header.vci)))
	      break;
	      
	      if (vcc)	/* found a suitable vcc */
		{
		  struct sk_buff* skb;
		  int pdulen = 0;
		  
		  for (j=0; j<rpd->nseg; j++)
		    pdulen += rpd->segment[j].length;
		  
		  /* pca200e->rx_queue = rxq;  for fetch */
		  skb = vcc->peek(vcc, pdulen, NULL);
		  
		  if (skb)
		    {
		      skb->free = 1;
		      
		      for (j=0; j<rpd->nseg; j++)
			memcpy(skb_put(skb, rpd->segment[j].length),
			       ((qhost_rbuff_t*) bus_to_virt(rpd->segment[j].handle))->buff,
			       rpd->segment[j].length);
		      
#if CONFIG_ATM_PCA200E_DEBUG > 3
		      {
			int ii;
			
			/* trace data PDUs only */
			if (vcc->vci >= ATM_NOT_RSV_VCI)
			  {
			    printk("intrx: rxpdu: %d.%d.%d:%d len=%d\n",
				   vcc->itf, vcc->vpi, vcc->vci,
				   def2aal(vcc->aal), pdulen);
			    for (ii=0; ii<skb->len; ii++)
			      printk("%02x ", ((unsigned char*)skb->data)[ii]);
			    printk("\n");
			  };
		      };
#endif
		      
		      vcc->push(vcc, skb);
		      vcc->stats->rx++;
		    }
		  else	/* invalid skb */
		    vcc->stats->rx_drop++;
		}
	      else	/* no suitable vcc found */
		{
		/* This shouldnt happen. The card should deliver only PDUs
		   for VCIs opened before */
		/* log this to aal0 stats as we cant figure out the aal now */
		dev->stats.aal0.rx_drop++;
#if CONFIG_ATM_PCA200E_DEBUG > 3
		DPRINTK("no vcc found for %d.%d.%d\n",
			dev->number, rpd->atm_header.vpi, rpd->atm_header.vci);
#endif			
		}
	    }
	  else	/* pdu had error bit set */
	    {
	      /* log this to aal0 stats as we cant figure out the aal now */
	      dev->stats.aal0.rx_err++;
#if CONFIG_ATM_PCA200E_DEBUG > 3
		DPRINTK("damaged PDU on %d.%d.%d\n",
			dev->number, rpd->atm_header.vpi, rpd->atm_header.vci);
#endif			
	    };

	  /* put the buffers into the free list */
	  for (j=0; j<rpd->nseg; j++)
	    {
	      switch (((qhost_rbuff_t*) bus_to_virt(rpd->segment[j].handle))->type)
		{
		case bmag_small: 
		  {
		    pca200e->smlb1_count--;
		    ((qhost_rbuff_t*) bus_to_virt(rpd->segment[j].handle))->next = pca200e->smlb1_rbuff_head;
		    pca200e->smlb1_rbuff_head = (qhost_rbuff_t*) bus_to_virt(rpd->segment[j].handle);
		  }; break;
		case bmag_large: 
		  {
		    pca200e->lrgb1_count--;
		    ((qhost_rbuff_t*) bus_to_virt(rpd->segment[j].handle))->next = pca200e->lrgb1_rbuff_head;
		    pca200e->lrgb1_rbuff_head = bus_to_virt(rpd->segment[j].handle);
		  }; break;
		};
	    };
	  if (pca200e->smlb1_count <= SMALL_B1_LOW_WATER)
	    pca200e_supply(pca200e, 1);
	  if (pca200e->lrgb1_count <= LARGE_B1_LOW_WATER)
	    pca200e_supply(pca200e, 0);
	  writel((unsigned long) rxq->rpd, &rxq->qentry->ioblock);
	  SET_STATUS(rxq, stat_pending);
	}
      else /* no more used entries */
	break;
      rxq = rxq->next;
    };
  pca200e->rx_queue = rxq;
};


static int
pca200e_send(struct atm_vcc *vcc, struct sk_buff *skb)
{
  qhost_tx_t*	txq;
  tpd_t*	tpd;
  unsigned long flg;

  /* here begins a critical section */
  save_flags(flg);
  cli();

  txq = PCA200E_DEV(vcc->dev)->tx_queue;
  tpd = txq->tpd;

  if (!STAT_FREE(txq))
    {
      pca200e_inttx(vcc->dev);	/* try to free completed txqueue entries */
      if (!STAT_FREE(txq))
	{
	  restore_flags(flg);
	  vcc->stats->tx_err++;
	  printk(DEV_LABEL "TX problem: itf=%d, queue saturated -> PDU dropped\n", vcc->itf);
	  return -EBUSY;
	};
    };

  if (!(((unsigned long)(skb->data)) & 0x00000003))
    tpd->segment[0].buffer = virt_to_bus(skb->data);
  else
    {
      DPRINTK("pca200e_send: misaligned data detected.\n");
      txq->txbuff = kmalloc(skb->len, GFP_ATOMIC);
      if (!(txq->txbuff))
	{
	  restore_flags(flg);
	  return -ENOMEM;
	};
      memcpy(txq->txbuff, skb->data, skb->len);
      tpd->segment[0].buffer = virt_to_bus(txq->txbuff);
    };

  PCA200E_DEV(vcc->dev)->tx_queue = txq->next;
  restore_flags(flg);
  /* here ends the critical section */

#if CONFIG_ATM_PCA200E_DEBUG > 3
#if CONFIG_ATM_PCA200E_DEBUG < 5
  if (vcc->vci >= ATM_NOT_RSV_VCI)
#endif
    {
      int i;
      printk("pca200e_send: %d.%d.%d:%d head=%p, data=%p, len=%lu\n",
	      vcc->itf, vcc->vpi, vcc->vci, def2aal(vcc->aal),
	      skb->head, skb->data, skb->len);
      for (i=0; i<skb->len; i++)
	printk("%02x ", ((unsigned char*)skb->data)[i]);
      printk("\n");
    };
#endif

  txq->vcc = vcc;
  txq->skb = skb;

  tpd->rc_stream_desc.empty_cells = 0x0000;
  if ((vcc->qos.txtp.max_pcr > 0) && (vcc->qos.txtp.max_pcr != ATM_OC3_PCR))
    {
      /* calculate index in rate table from tx-rate - XXX recheck with docs */
      tpd->rc_stream_desc.data_cells = vcc->qos.txtp.max_pcr*255/353207+1;
#if 0
      printk(DEV_LABEL
	     "send: %d.%d.%d:%d,"
	     "txtp.max_pcr=%d, rxtp.max_pcr=%d -> entry %d\n",
	     vcc->itf, vcc->vpi, vcc->vci, def2aal(vcc->aal),
	     vcc->qos.txtp.max_pcr, vcc->qos.rxtp.max_pcr,
	     tpd->rc_stream_desc.data_cells);
#endif
    }
  else
    tpd->rc_stream_desc.data_cells = 0x0000;

  tpd->atm_header.clp = 0;
  tpd->atm_header.plt = 0;
  tpd->atm_header.vci = vcc->vci;
  tpd->atm_header.vpi = vcc->vpi;
  tpd->atm_header.gfc = 0;

  tpd->spec.pdu_len = skb->len;
  tpd->spec.num_seg = 1;
  tpd->spec.aal	    = def2aal(vcc->aal);
#ifdef PCA200E_SYNC_SEND
  tpd->spec.intr    = 0;
#else
  tpd->spec.intr    = 1;
#endif
  tpd->segment[0].length = ((skb->len + 3) & 0x1fffc);

#if !defined(SEND_BENCH)           /* set this to 0 for protocol stack measurements */

  SET_STATUS(txq, stat_pending);
  writel((unsigned long) tpd+TPD_EXTENSION+1, &txq->qentry->ioblock);

#ifdef PCA200E_SYNC_SEND
{
  int i;
  i = POLL_STAT_COMPLETE(txq, 2);
  SET_STATUS(txq, stat_free);
  if (((unsigned long) skb->data) & 0x00000003)
    kfree(txq->txbuff);
  vcc->pop(vcc, skb);
  return i;
};
#endif

#else	/* SEND_BENCH */
  vcc->pop(vcc, skb);
#endif
  TRACE;
  return 0;
};


int
pca200e_activatevcin(struct pca200e_dev* pca200e,
		     unsigned int activate,
		     unsigned int aal, unsigned int vpi, unsigned int vci,
		     unsigned int mtu)
     /* tries to [de]activate the specified aal/vpi/vci
	activate == TRUE  -> activate
	activate == FALSE -> deactivate
	returns TRUE on success */
{
  int i;
  qhost_cmd_t*	cq;
  
  TRACE;
  cq = pca200e->cmd_queue;
  /* vpi has to be zero */
  writel((unsigned int) vci, &cq->qentry->cmd.activate_param.vpvc);

  SET_STATUS(cq, stat_pending);
  writel((activate ? op_activate_vcin : op_deactivate_vcin) | (aal << 8),
	 &cq->qentry->cmd.activate_param.op);

  i = POLL_STAT_COMPLETE(cq, PCA200E_COMMAND_TIMEOUT);

  SET_STATUS(cq,stat_free);
  pca200e->cmd_queue = cq->next;

  return !i;
};




void
pca200e_interrupt(int irq, struct atm_dev* dev, struct pt_regs *regs)
{
  TRACE;
  /* check if we caused this interrupt */
  if (0x01 != readb(PCA200E_DEV(dev)->virt_rambase + PCA200E_PSR_OFFSET))
    {
      printk(DEV_LABEL "Unexpected Interrupt %d\n", irq);
      return;
    };
  pca200e_intrx(dev);
#ifndef PCA200E_SYNC_SEND
  pca200e_inttx(dev);
#endif

  /* signalling interrupt acknowledge */
  writeb(0x10, PCA200E_DEV(dev)->virt_rambase + PCA200E_HCR_OFFSET);
};


static int
pca200e_init(struct atm_dev* dev)
{
  if (!PCA200E_DEV(dev))
    return -EFAULT;
  
  if (pca200e_map(PCA200E_DEV(dev))) return -ENOMEM;
  if (pca200e_reset(PCA200E_DEV(dev))) return -ENOMEM;
  pca200e_loadfirm(PCA200E_DEV(dev));
  if (pca200e_startfirm(PCA200E_DEV(dev))) return -ENOMEM;
  if (pca200e_initialize(PCA200E_DEV(dev))) return -ENOMEM;
  if (pca200e_initcmdqueue(PCA200E_DEV(dev))) return -ENOMEM;
  if (pca200e_inittxqueue(PCA200E_DEV(dev))) return -ENOMEM;
  if (pca200e_initrxqueue(PCA200E_DEV(dev))) return -ENOMEM;

  if (pca200e_initbuffqueue(PCA200E_DEV(dev), 1)) return -ENOMEM;
  if (pca200e_initbuffqueue(PCA200E_DEV(dev), 0)) return -ENOMEM;
  
  if (pca200e_readprom(PCA200E_DEV(dev))) return -EFAULT;
  
  if(request_irq(PCA200E_DEV(dev)->irq,
		 (void*)pca200e_interrupt,
		 SA_INTERRUPT,
		 PCA200E_DEV(dev)->name,
		 dev))
    {
      printk(DEV_LABEL "unable to reserve IRQ %d for %s\n",
	     PCA200E_DEV(dev)->irq, PCA200E_DEV(dev)->name);
      return -EBUSY;
    };

  PCA200E_DEV(dev)->cardstate = cs_int;
  PCA200E_DEV(dev)->available_cell_rate = ATM_OC3_PCR;
  dev->ci_range.vpi_bits = 8;
  dev->ci_range.vci_bits = 10;
  return 0;
};


/*-----------------------------------------------*/
static int
pca200e_open(struct atm_vcc *vcc, short vpi, int vci)
{
#ifdef CONFIG_ATM_PCA200E_DEBUG
  char* trafclass[] = {"", "NONE", "UBR", "CBR", "VBR", "ABR", "ANY"};
#endif CONFIG_ATM_PCA200E_DEBUG

  if (vci == ATM_VPI_ANY) /* look for a free VPI */
    {
      struct atm_vcc* walk = vcc->dev->vccs;
      vpi = 0;
      for (walk = vcc->dev->vccs; walk; walk=walk->next)
	{
	  if ((walk->vci == vci) && (walk->vpi == vpi))
	    {
	      vpi++;
	      walk = vcc->dev->vccs;	/* and start over again */
	    };
	}
    };
  
  if (vci == ATM_VCI_ANY) /* look for a free VCI */
    {
      struct atm_vcc* walk = vcc->dev->vccs;
      vci = ATM_NOT_RSV_VCI;
      for (walk = vcc->dev->vccs; walk; walk=walk->next)
	{
	  if ((walk->vpi == vpi) && (walk->vci == vci))
	    {
	      vci = walk->vci + 1;
	      walk = vcc->dev->vccs;	/* and start over again */
	    };
	}
    };
  if (vci != ATM_VPI_UNSPEC && vpi != ATM_VCI_UNSPEC)
    vcc->flags |= ATM_VF_ADDR;
  vcc->vpi = vpi;
  vcc->vci = vci;

  /* happily return if this was for resource checking only */
  if (vci == ATM_VPI_UNSPEC || vpi == ATM_VCI_UNSPEC) return 0;

  vcc->itf = vcc->dev->number;

  DPRINTK(DEV_LABEL __FUNCTION__
	  ": %d.%d.%d:%d QoS(TX: cl=%s, pcr=%d-%d, cdv=%d, sdu=%d; "
	  "RX: cl=%s, pcr=%d-%d, cdv=%d, sdu=%d)\n",
	  vcc->itf, vcc->vpi, vcc->vci, def2aal(vcc->aal),
	  trafclass[vcc->qos.txtp.traffic_class],
	  vcc->qos.txtp.min_pcr, vcc->qos.txtp.max_pcr,
	  vcc->qos.txtp.max_cdv, vcc->qos.txtp.max_sdu,
	  trafclass[vcc->qos.rxtp.traffic_class],
	  vcc->qos.rxtp.min_pcr, vcc->qos.rxtp.max_pcr,
	  vcc->qos.rxtp.max_cdv, vcc->qos.rxtp.max_sdu);

  /* check available bandwidth for non-UBR connections */
  if (vcc->qos.txtp.traffic_class != ATM_UBR)
    if (vcc->qos.txtp.max_pcr > 0)
      if (PCA200E_DEV(vcc->dev)->available_cell_rate < vcc->qos.txtp.max_pcr)
	return -EAGAIN;

  if (!pca200e_activatevcin(PCA200E_DEV(vcc->dev),
			    1 /* TRUE */,
			    def2aal(vcc->aal), vcc->vpi, vcc->vci,
			    48))
    return -EBUSY;

  /* if non-UBR conn, decrease available bandwidth for non-UBR connections */
  if (vcc->qos.txtp.traffic_class != ATM_UBR)
    if (vcc->qos.txtp.max_pcr > 0)
      PCA200E_DEV(vcc->dev)->available_cell_rate -= vcc->qos.txtp.max_pcr;
  
#ifdef MODULE
  MOD_INC_USE_COUNT;
#endif
  vcc->flags |= ATM_VF_READY;

  return 0;
};


static void
pca200e_close(struct atm_vcc *vcc)
{
  DPRINTK(DEV_LABEL __FUNCTION__": %d.%d.%d:%d\n",
	 vcc->itf, vcc->vpi, vcc->vci, def2aal(vcc->aal));
  if (pca200e_activatevcin(PCA200E_DEV(vcc->dev),
			   0 /* FALSE */,
			   def2aal(vcc->aal), vcc->vpi, vcc->vci,
			   0))
    {
#ifdef MODULE
      MOD_DEC_USE_COUNT;
#endif
    };

  /* if non-UBR conn, increase available bandwidth for non-UBR connections */
  if (vcc->qos.txtp.traffic_class != ATM_UBR)
    if (vcc->qos.txtp.max_pcr > 0)
      PCA200E_DEV(vcc->dev)->available_cell_rate += vcc->qos.txtp.max_pcr;
};


unsigned int
dmaswap(unsigned int in)
{
#ifdef __i386__
  asm("bswap %%eax\n"
      : "=a" (in)
      : "a" (in));
  return in;
#endif
#ifdef __alpha__
  return in;
#endif
};

static int
pca200e_ioctl(struct atm_dev *dev,unsigned int cmd, unsigned long arg)
{
  switch (cmd)
    {
    case SONET_GETSTATZ:
    case SONET_GETSTAT:
      {
	struct sonet_stats sost;
	stats200_t* stat;
	
	stat = (stats200_t*) kmalloc(sizeof(stats200_t), GFP_KERNEL);
	if (!stat)
	  return -ENOMEM;

	memset((void*) stat, 0, sizeof(stats200_t));
	pca200e_getstats(PCA200E_DEV(dev), stat);
	sost.section_bip = dmaswap(stat->phyoc3.section_bip8_errors);
	sost.line_bip	 = dmaswap(stat->phyoc3.line_bip24_errors);
	sost.path_bip	 = dmaswap(stat->phyoc3.path_bip8_errors);
	sost.line_febe	 = dmaswap(stat->phyoc3.line_febe_errors);
	sost.path_febe	 = dmaswap(stat->phyoc3.path_febe_errors);
	sost.corr_hcs	 = dmaswap(stat->phyoc3.corr_hcs_errors);
	sost.uncorr_hcs	 = dmaswap(stat->phyoc3.ucorr_hcs_errors);
	sost.tx_cells	 = (dmaswap(stat->atm.cells_transmitted)
			    +dmaswap(stat->aal0.cells_transmitted)
			    +dmaswap(stat->aal4.cells_transmitted)
			    +dmaswap(stat->aal5.cells_transmitted));
	sost.rx_cells	 = (dmaswap(stat->atm.cells_received)
			    +dmaswap(stat->aal0.cells_received)
			    +dmaswap(stat->aal4.cells_received)
			    +dmaswap(stat->aal5.cells_received));
	memcpy_tofs((struct sonet_stats*)arg, &sost,
		    sizeof(struct sonet_stats));
	kfree((void*)stat);
	break;
      }
    case SONET_SETDIAG:
    case SONET_CLRDIAG:
    case SONET_GETDIAG:
      return 0; break;
    case PCA200E_DUMP:
      {
	stats200_t* stat;
	
	memcpy_tofs((char*)(&(((struct pca200e_dump*) arg)->dev)),
		    (char*)PCA200E_DEV(dev),
		    sizeof(struct pca200e_dev));
	memcpy_tofs((char*)(&(((struct pca200e_dump*) arg)->cpqueue)),
		    (char*)(PCA200E_DEV(dev)->cpqueue),
		    sizeof(cp_queues_t));
	memcpy_tofs((char*)(&(((struct pca200e_dump*) arg)->monitor)),
		    (char*)(PCA200E_DEV(dev)->virt_rambase + PCA200E_MON960_OFFSET),
		    sizeof(monitor_t));
	stat = (stats200_t*) kmalloc(sizeof(stats200_t), GFP_KERNEL);

	if (stat)
	  {
	    pca200e_getstats(PCA200E_DEV(dev), stat);
	    memcpy_tofs((void*) &(((struct pca200e_dump*) arg)->stats),
			(void*) stat,
			sizeof(stats200_t));
	    kfree((void*)stat);
	  };
	return 0;
      };
    default:
      DPRINTK(">pca200e_ioctl: cmd=%08x, arg=%08lx\n",
	      cmd, arg);
    };
  return 0;
};


static const struct atmdev_ops ops =
{
  pca200e_open,
  pca200e_close,
  pca200e_ioctl,
  NULL, /*pca200e_getsockopt,*/
  NULL, /*pca200e_setsockopt,*/
  pca200e_send,
  NULL, /*pca200e_sg_send,*/
  NULL,	/*pca200e_poll,*/
  NULL, /*pca200e_send_oam,*/
  NULL, /*pca200e_phy_put,*/
  NULL, /*pca200e_phy_get,*/
  NULL, /*pca200e_feedback,*/
  NULL, /*pca200e_change_qos,*/
  NULL  /*pca200e_free_rx_skb*/
};


int
pca200e_detect(void)
{
  struct atm_dev*	dev;
  struct pca200e_dev*	pca200e;
  int			index;
  int			i;
  
  pca200e = (struct pca200e_dev *) kmalloc(sizeof(struct pca200e_dev),
					   GFP_KERNEL);
  if (!pca200e)
    {
      return -ENOMEM;
    };
  memset(pca200e, 0, sizeof(struct pca200e_dev));
  pca200e->tx_wait_queue = NULL;

  index = 0;
  while (!pca200e_card_detect(pca200e, index))
    {
      (unsigned long) pca200e->bufferlist[0] = 1;
      sprintf(pca200e->name, "PCA-200E-%d",index);
      printk(DEV_LABEL "found FORE PCA-200E at 0x%08lx, IRQ %d\n",
	     pca200e->phys_rambase, pca200e->irq);
      dev = atm_dev_register(DEV_LABEL, &ops, 0); 
      if (!dev) break;
      PCA200E_DEV(dev) = pca200e;
      if (pca200e_init(dev))
	{
	  atm_dev_deregister(dev);
	  break;
	}
      for (i=0; i<ESI_LEN; i++)
	dev->esi[i] = PCA200E_DEV(dev)->mac[i];
      dev->type = "pca200e";
      pca200e->next = pca200e_boards;
      pca200e_boards = pca200e;
      pca200e->atmdev = dev;
      index++;
      pca200e = (struct pca200e_dev *) kmalloc(sizeof(struct pca200e_dev),
					       GFP_KERNEL);
      if (!pca200e) break;
      memset(pca200e, 0, sizeof(struct pca200e_dev));
//break; /* uncomment this to avoid more than one card gets detected */ 
    };
  if (pca200e)
    kfree(pca200e);
  return index;
};


#ifdef MODULE
unsigned int init_module(void)
{
  int i;
  printk(DEV_LABEL "FORE Systems PCA-200E driver v0.3 - Uwe.Dannowski@inf.tu-dresden.de\n");
  i = (pca200e_detect() == 0);
  return i;
};

void
pca200e_deinit(struct pca200e_dev* pca200e)
{
  unsigned long* ip;
  
  switch (pca200e->cardstate)
    {
    case cs_complete:	/* will never be reached */
    case cs_int:		/* free associated interrupt */
      free_irq(pca200e->irq, pca200e->atmdev);
      
    case cs_init_lb1:
    case cs_init_sb1:
      
    case cs_init_rx:
      
      /*
       * smlb1_queue and lrgb1_queue have to be freeeeeeed, too.
       */
      
      if (pca200e->smlb1_status_page)
	free_pages((long) pca200e->smlb1_status_page,
		   (QSIZE_RESUPPLY*32+4095)/4096);
      if (pca200e->lrgb1_status_page)
	free_pages((long) pca200e->lrgb1_status_page,
		   (QSIZE_RESUPPLY*32+4095)/4096);
      if (pca200e->smlb1_rbd_page)
	free_pages((long) pca200e->smlb1_rbd_page,
		   (QSIZE_RESUPPLY*RBD_BLK_SIZE*sizeof(rbd_t)+4095)/4096);
      if (pca200e->lrgb1_rbd_page)
	free_pages((long) pca200e->lrgb1_rbd_page,
		   (QSIZE_RESUPPLY*RBD_BLK_SIZE*sizeof(rbd_t)+4095)/4096);
      
      ip = (unsigned long*) pca200e->bufferlist;
      while(--*ip)
	{
	  if (pca200e->bufferlist[*ip])
	    {
	      if (pca200e->bufferlist[*ip]->buff)
		kfree(pca200e->bufferlist[*ip]->buff);
	      kfree((void*) (pca200e->bufferlist[*ip]));
	    };
	};
      
    case cs_init_tx:		/* undo all before cs_init_rx */
      if (pca200e->rx_queue)
	{
	  qhost_rx_t* tmp1 = pca200e->rx_queue->next;
	  pca200e->rx_queue->next = NULL;
	  while (tmp1)
	    {
	      pca200e->rx_queue = tmp1;
	      tmp1 = tmp1->next;
	      kfree((void*) pca200e->rx_queue);
	    };
	};
      if (pca200e->rx_status_page)
	free_pages((long) pca200e->rx_status_page,
		   (QSIZE_RX*32+4095)/4096);
      if (pca200e->rpd_page)
	free_pages((long) pca200e->rpd_page,
		   (QSIZE_RX*RPD_SIZE+4095)/4096);
      
    case cs_init_cmd:		/* undo all before cs_init_tx */
      if (pca200e->tx_queue)
	{
	  qhost_tx_t* tmp1 = pca200e->tx_queue->next;
	  pca200e->tx_queue->next = NULL;
	  while (tmp1)
	    {
	      pca200e->tx_queue = tmp1;
	      tmp1 = tmp1->next;
	      kfree((void*) pca200e->tx_queue);
	    };
	};
      if (pca200e->tx_status_page)
	free_pages((long) pca200e->tx_status_page,
		   (QSIZE_TX*32+4095)/4096);
      if (pca200e->tpd_page)
	free_pages((long) pca200e->tpd_page,
		   (QSIZE_TX*TPD_SIZE+4095)/4096);
      
    case cs_init:		/* undo all before cs_init_cmd */
      if (pca200e->cmd_queue)
	{
	  qhost_cmd_t* tmp1 = pca200e->cmd_queue->next;
	  pca200e->cmd_queue->next = NULL;
	  while (tmp1)
	    {
	      pca200e->cmd_queue = tmp1;
	      tmp1 = tmp1->next;
	      kfree((void*) pca200e->cmd_queue);
	    };
	};
      if (pca200e->cmd_status_page)
	free_pages((long) pca200e->cmd_status_page,
		   (QSIZE_CMD*32+4095)/4096);
      
    case cs_startfirm:	/* reset the card */
      pca200e_reset(pca200e);
      
    case cs_loadfirm:	/* invalidate firmware */
      writel(0, pca200e->virt_rambase
	     + _binary_pca200e_bin_start->startoffset); 
    case cs_reset:
      
    case cs_map:		/* release shared memory mapping */
#ifndef __alpha__
      iounmap((void*) pca200e->virt_rambase);
#endif
      
    case cs_blank:
    };
};

void
cleanup_module(void)
{
  struct pca200e_dev* tmp1;
  while (pca200e_boards)
    {
      atm_dev_deregister(pca200e_boards->atmdev);
      pca200e_deinit(pca200e_boards);
      tmp1 = pca200e_boards;
      pca200e_boards = pca200e_boards->next;
      kfree(tmp1);
    };
  printk("pca200e removed\n");
};

#endif



